searchNodes=[{"doc":"This task prints routing between events and handlers setup by Trolleybus . Usage mix trolleybus . routes","ref":"Mix.Tasks.Trolleybus.Routes.html","title":"Mix.Tasks.Trolleybus.Routes","type":"task"},{"doc":"Defines a local, application-level PubSub API for dispatching side effects. This PubSub mechanism is dedicated to handling side-effects in business logic. Instead of calling side effects directly, an event is published. The event is then routed to one or more handlers, according to declared routing. Example Let's assume we have a code path in business logic where we want to trigger two side effects after core logic is done. def accept_invite ( ... ) do ... App.Emails . notify_invite_accepted ( document , inviter , user ) App.Webhooks . notify_auth_update ( document , user ) ... end First, we define the event using Trolleybus.Event : defmodule App.Documents.Events.MembershipInviteAccepted do use Trolleybus.Event handler App.Documents.EmailEventHandler handler App.Webhooks.EventHandler message do field :membership , % App.Memberships.Membership { } field :document , % App.Documents.Document { } field :inviter , % App.Users.User { } , required : false field :user , % App.Users.User { } end end Next, we define new or extend existing event handlers using Trolleybus.Handler : defmodule App.Documents.EmailEventHandler do use Trolleybus.Handler alias App.Documents.Events.MembershipInviteAccepted def handle_event ( % MembershipInviteAccepted { document : document , inviter : inviter , user : user } ) do App.Emails . notify_invite_accepted ( document , inviter , user ) end ... end defmodule App.Webhooks.EventHandler do use Trolleybus.Handler alias App.Documents.Events.MembershipInviteAccepted def handle_event ( % MembershipInviteAccepted { document : document , user : user } ) do App.Webhooks . notify_auth_update ( document , user ) end ... end Finally, we publish an event instead of calling Emails and Webhooks directly: alias App.Documents.Events def accept_invite ( ... ) do ... Trolleybus . publish ( % Events.MembershipInviteAccepted { membership : membership , document : document , inviter : inviter , user : user } ) ... end Publishing A publish/2 call triggers the dispatch logic for the event. The dispatch retrieves all handlers provided in event's definition and passes it to those handlers. Publishing may be executed in three modes: fully synchronous :full_sync (default) - executes all handlers sequentially and synchronously, within the same process as caller, asynchronous :async - executes side effects in separate processes not waiting for them to finish executing, synchronous :sync - blocks until all processes executing event handlers complete. The wait time in case of synchronous publish is limited by timeout setup via :sync_timeout option, expressed in milliseconds (defaults to 5000). Any handler execution process running for longer than the setup timeout is killed. For simplicity, Trolleybus is neither concerned with persistence nor retrying failed handler calls. If a particular case calls for this kind of guarantees, it can be realised by, for instance, making relevant handler use job processing library like Oban . Buffering There are cases where publishing events may have to be either deferred or abandoned completely. This problem can be solved by using one of buffering wrappers: transaction/1 , buffered/1 or muffled/1 . The underlying mechanism is the same for all three. Before a block of code is executed, a buffer is opened. The buffer is an agent process storing a list of events along with their respective publish options, if any are passed. Each call to publish/2 inside that block puts the event in that list instead of actually publishing it. Once the block of code finishes executing, further behavior depends on the wrapper. buffered/1 and muffled/1 discard the events (the former returns them too, along with the result). transaction/1 fetches contents of the current buffer, closes it and re-publishes all the events from the list. If wrappers are nested, the re-publishing will result by putting the events in another buffer opened by the outer wrapper. Eventually, the events get either dispatched to respective handlers or discarded. Current buffer as well as buffer stack are tracked using process dictionary. This means that buffering will only work for code executed within a single process. Events inside transactions In case of more complex logic, there can be multiple events published along a single code path. This code path can in turn contain multiple subroutines changing system state, wrapped in one big transaction. It's often hard to defer all the side effects outside of that transaction without clunky workarounds. That's where transaction/1 comes into play. It's especially useful when events are published inside a large, multi-stage Ecto.Multi based transaction. Assuming we have events published somewhere inside a big transaction, like this: def update_memberships ( memberships ) do Ecto.Multi . run ( :updated_memberships , fn _ , %{ old_memberships : memberships } -&gt; updated_memberships = Enum . map ( memberships , fn membership -&gt; Trolleybus . publish ( % MembershipChanged { ... } ) ... end ) { :ok , updated_memberships } end ) end we can defer actual publishing of all events until after the transaction is executed by wrapping the top level Ecto transaction with Trolleybus transaction: def process ( multi ) do case Trolleybus . transaction ( fn -&gt; Repo . transaction ( multi ) end ) do { :ok , ... } -&gt; ... end end Nesting transactions Buffered transactions wrapped with transaction/1 can be arbitrarily nested and it's guaranteed that only the outermost one will publish the events: Trolleybus . transaction ( fn -&gt; # all events will be published # only after this outer block finishes ... Trolleybus . transaction ( fn -&gt; ... end ) ... end ) Reusing code publishing events We often want to reuse existing logic as a part of larger, more complex routines. The problem is that this existing logic may already publish events specific to that original context. It may often be undesirable, because we want to emit events specific to the wrapping routine and need existing logic only for its data manipulation part. Another use case where that may be an issue is reusing business logic for setting up system state in tests instead of synthetic factories. In order to make it possible, we can wrap the reused code with muffled/1 : { :ok , %{ membership : accepted_membership } } = Trolleybus . muffled ( fn -&gt; AcceptInvite . accept_membership_invite ( user , membership ) end ) Wrapping it this way will send all the events to a buffer which will be promptly discarded after the code block completes. Testing events When we want to test what events does a given piece of logic publish instead of relying on checking side effects, we can use buffered/1 : { { :ok , %{ membership } } , events } = Trolleybus . buffered ( fn -&gt; AcceptInvite . accept_membership_invite ( user , membership ) end ) assert [ { % MembershipInviteAccepted { } , [ ] } ] = events Events published inside buffered/1 are stored in a buffer, whose contents are returned along with the result of the code block in a tuple. The events are then promptly discarded, so no handler gets triggered. Another potentially handy function is get_buffer/0 , which allows to &quot;peek&quot; into contents of the buffer at any point when buffer is open. It can be used inside any of buffered/1 , muffled/1 or transaction/1 blocks: Trolleybus . muffled ( fn -&gt; ... assert [ { % SomeEvent { } , [ ] } ] = get_buffer ( ) ... end ) One important thing to note is that publishing mode can be overridden in all calls using Application configuration: config :trolleybus , mode_override : :full_sync This allows users to force all events to be published using a given mode, for example :full_sync in config/test.exs to make testing side effects simpler. This lets us avoiding any issues related to running handlers in a separate process, like having to explicitly handle Ecto sandbox allowance. Listing routes In order to print all events and associated handlers in the project, a dedicated mix task can be run: mix trolleybus . routes The output has a following form: * App.Events.DocumentTransferred =&gt; App.Webhooks.EventHandler =&gt; App.Memberships.EmailEventHandler * App.Events.UserInvitedToDocument =&gt; App.Memberships.EmailEventHandler ...","ref":"Trolleybus.html","title":"Trolleybus","type":"module"},{"doc":"Returns events published in the given function. The events are returned along with the result of the function, without dispatching them to defined handlers. Order of events in the list is always consistent with order of publishing inside the wrapped function. Example { &quot;result&quot; , [ { % SomeEvent { } , [ ] } , { % OtherEvent { } , mode : :async } , { % AnotherEvent { } , mode : :sync , sync_timeout : 1_000 } ] } = Trolleybus . buffered ( fn -&gt; Trolleybus . publish ( % SomeEvent { name : &quot;value&quot; } ) Trolleybus . publish ( % OtherEvent { flag : true } , mode : :async ) Trolleybus . publish ( % AnotherEvent { number : 123 } , mode : :sync , sync_timeout : 1_000 ) &quot;result&quot; end )","ref":"Trolleybus.html#buffered/1","title":"Trolleybus.buffered/1","type":"function"},{"doc":"Lists all published but not yet dispatched events so far. Order of events in the list is always consistent with order of publishing inside the wrapped function. Example Trolleybus . muffled ( fn -&gt; Trolleybus . publish ( % SomeEvent { name : &quot;value&quot; } ) Trolleybus . publish ( % OtherEvent { flag : true } , mode : :async ) [ { % SomeEvent { } , [ ] } , { % OtherEvent { } , mode : :async } ] = Trolleybus . get_buffer ( ) Trolleybus . publish ( % AnotherEvent { number : 123 } , mode : :sync , sync_timeout : 1_000 ) [ { % SomeEvent { } , [ ] } , { % OtherEvent { } , mode : :async } , { % AnotherEvent { } , mode : :sync , sync_timeout : 1_000 } ] = Trolleybus . get_buffer ( ) end )","ref":"Trolleybus.html#get_buffer/0","title":"Trolleybus.get_buffer/0","type":"function"},{"doc":"Discards events published in the given function without dispatching them to defined handlers. Example &quot;result&quot; = Trolleybus . muffled ( fn -&gt; Trolleybus . publish ( % SomeEvent { name : &quot;value&quot; } ) &quot;result&quot; end )","ref":"Trolleybus.html#muffled/1","title":"Trolleybus.muffled/1","type":"function"},{"doc":"Publishes event. Example :ok = Trolleybus . publish ( % SomeEvent { name : &quot;value&quot; } ) :ok = Trolleybus . publish ( % OtherEvent { flag : true } , mode : :async ) :ok = Trolleybus . publish ( % AnotherEvent { number : 123 } , mode : :sync , sync_timeout : 1_000 ) Options :mode - Event dispatch mode. Can be one of :full_sync , :async or :sync . See &quot;Dispatch modes&quot; below for detailed explanation. Default: :full_sync . :sync_timeout - Timeout in milliseconds after which synchronous dispatch is cancelled and handler processes still in progress are killed. Works only for :sync mode. Default: 5000 Disptach modes Publishing may be executed in three modes: :full_sync - Executes all handlers sequentially and synchronously, within the same process as caller. :async - Executes handlers in separate processes fully asynchronously, not waiting for them to finish executing. :sync - Executes handlers in separate processes in parallel and waits until they complete executing. Waiting time is determined by :sync_timeout option. Handler processes running past timeout are killed.","ref":"Trolleybus.html#publish/2","title":"Trolleybus.publish/2","type":"function"},{"doc":"Dispatches published events after running the given function. Dispatching to event handlers is done only if the result is in {:ok, ...} tuple format. Otherwise, events are discarded. Example { :ok , &quot;result&quot; } = Trolleybus . transaction ( fn -&gt; Trolleybus . publish ( % SomeEvent { name : &quot;value&quot; } ) { :ok , &quot;result&quot; } end )","ref":"Trolleybus.html#transaction/1","title":"Trolleybus.transaction/1","type":"function"},{"doc":"","ref":"Trolleybus.html#t:publish_mode/0","title":"Trolleybus.publish_mode/0","type":"type"},{"doc":"","ref":"Trolleybus.html#t:publish_option/0","title":"Trolleybus.publish_option/0","type":"type"},{"doc":"Application definition for Trolleybus. Starts a Task.Supervisor instance for running tasks executing handlers when publishing events.","ref":"Trolleybus.Application.html","title":"Trolleybus.Application","type":"module"},{"doc":"Defines an event struct for publishing via Trolleybus . An event definition is composed of a couple of elements: handler/1 declarations - they define which handlers are called when event is published. message/1 declaration - determines the underlying struct and types of fields using field/3 validated before publishing. Example defmodule App.Events.EmailInvitedToDocument do use Trolleybus.Event handler ( App.Handlers.EmailHandler ) handler ( App.Handlers.StripeHandler ) message do field ( :invitee_email , App.Types.Email ) field ( :document , % Document { } ) field ( :inviter , % User { } ) field ( :message , :string , required : false ) field ( :vat_invoice? , :boolean , default : false ) end end Event can declare any number of handlers, including zero. Message shape is declared by a series of field/3 macro calls. Each field/3 declaration maps to a field in the underlying, generated struct. Events can be instantiated like any other struct: iex&gt; event = % App.Events.EmailInvitedToDocument { invitee_email : &quot;alice@example.com&quot; , document : document , inviter : user } Each event implements cast!/1 function which is called before publishing it. This function returns the same event struct with casted and validated message parameters as well as declared handlers.","ref":"Trolleybus.Event.html","title":"Trolleybus.Event","type":"module"},{"doc":"Defines a field on the event with a given name and type. All field definitions must be put inside message/1 block. The macro accepts three arguments: name - name of the field in the underlying struct. type - declares the expected type of the field value. It's validated iternally via cast!/1 . opts - additional options for the field. Currently, required and default are accepted. Options :required - boolean determining whether field value can be left empty ( nil ). If field's required option is set to true and event is instantiated and casted with that field left empty, an error is raised. Defaults to: true . :default - sets field value to the provided default when field is left empty during instantating the event. Defaults to: nil . Field types The underlying validation logic uses Ecto.Changeset , so basically any type accepted by Ecto.Schema is also accepted in message field definition. There's a special case for passing structs - %StructModule{} . This is because we don't want to validate exact contents of the struct, only that the value passed is a struct and is of matching type.","ref":"Trolleybus.Event.html#field/3","title":"Trolleybus.Event.field/3","type":"macro"},{"doc":"Indicates a handler the event is dispatched to. The macro accepts a module name. That name is later validated when casting the event internally via cast!/1 . If the handler does not implement the Trolleybus.Handler behaviour or does not explicitly handle that particular event, an error is raised.","ref":"Trolleybus.Event.html#handler/1","title":"Trolleybus.Event.handler/1","type":"macro"},{"doc":"Defines an event struct. Each struct field is defined using field/3 macro.","ref":"Trolleybus.Event.html#message/1","title":"Trolleybus.Event.message/1","type":"macro"},{"doc":"Defines a handler for events published via Trolleybus An event handler is expected to implement handle_event/1 callback. Example defmodule App.Handlers.StripeHandler do use Trolleybus.Handler def handle_event ( % EmailInvitedToDocument { email : email } = event ) do ... end def handle_event ( % UserInvitedToDocument { user : user , document : document ) do ... end end Defining event handlers The shape of the function head for the callback is expected to meet certain criteria. Each implemented clause must have one of the following shapes: def handle_event ( % EventModule { ... } ) do ... end or: def handle_event ( % EventModule { ... } = event ) do ... end Matching on the contents of the event is in turn restricted to binding keys to variables only, which means that this clause is accepted: def handle_event ( % EventModule { field : value } ) do ... end while this clause will raise an error: def handle_event ( % EventModule { field : %{ } = value ) do ... end Any other pattern in any of the clauses is going to result in raising an error. This restriction allows to implicitly infer a list of events supported by the handler, which is later used when validating handlers declared for the published event in Trolleybus.Event . This also enforces consistent pattern matching on exact events across all implemented handlers and guarantees exhaustive matching.","ref":"Trolleybus.Handler.html","title":"Trolleybus.Handler","type":"behaviour"},{"doc":"Defines event handler for handled events. Each supported event is expected to be handled by a distinct function clause using pattern matching. For more information on expected implementation of the callback, see &quot;Defining event handlers&quot; above.","ref":"Trolleybus.Handler.html#c:handle_event/1","title":"Trolleybus.Handler.handle_event/1","type":"callback"},{"doc":"Error raised on either invalid event module configuration or event struct failing validation.","ref":"Trolleybus.Event.Error.html","title":"Trolleybus.Event.Error","type":"exception"},{"doc":"Error raised on invalid handler definition.","ref":"Trolleybus.Handler.Error.html","title":"Trolleybus.Handler.Error","type":"exception"},{"doc":"Local, application-level PubSub API for dispatching side effects. Full documentation can be found on Hex .","ref":"readme.html","title":"Trolleybus","type":"extras"},{"doc":"The package can be installed by adding trolleybus to your list of dependencies in mix.exs : def deps do [ { :trolleybus , &quot;~&gt; 0.1.0&quot; } ] end","ref":"readme.html#installation","title":"Trolleybus - Installation","type":"extras"},{"doc":"Clone the repo and fetch its dependencies: git clone https://github.com/sketch-hq/trolleybus.git cd trolleybus mix deps.get mix test","ref":"readme.html#running-tests","title":"Trolleybus - Running tests","type":"extras"},{"doc":"Elixir source code is released under MIT License. Check LICENSE file for more information.","ref":"readme.html#license","title":"Trolleybus - License","type":"extras"},{"doc":"","ref":"changelog.html","title":"CHANGELOG","type":"extras"},{"doc":"Initial release","ref":"changelog.html#v0-1-0-2022-11-08","title":"CHANGELOG - v0.1.0 (2022-11-08)","type":"extras"},{"doc":"MIT License Copyright (c) 2022 Sketch Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the &quot;Software&quot;), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions: The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software. THE SOFTWARE IS PROVIDED &quot;AS IS&quot;, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.","ref":"license.html","title":"LICENSE","type":"extras"}]